@@ -40,38 +40,43 @@ sub new {
 
 sub toProcess {
   my ($self, $doc) = @_;
-  return $doc->findnode('//ltx:bibliography'); }
+  return $doc->findnodes('//ltx:bibliography'); }
 
+# Whooo, this is turning into a confusing API!
+# Potentially multiple biblist's in each document
+# potentially each one split into multiple output documents (need that option!)
+# Inevitably duplicates?
 sub process {
-  my ($self, $doc, $bib) = @_;
+  my ($self, $doc, @bibs) = @_;
+  my @docs = ($doc);
+  foreach my $bib (@bibs) {
+    next if $doc->findnodes('.//ltx:bibitem', $bib);    # Already populated?
 
-  my @bib = ($doc);
-  return $doc if $doc->findnodes('//ltx:bibitem', $bib);    # Already populated?
+    local $LaTeXML::Post::MakeBibliography::NUMBER = 0;
+    local %LaTeXML::Post::MakeBibliography::STYLE =
+      (map { ($_ => $bib->getAttribute($_)) } qw(bibstyle citestyle sort));
 
-  local $LaTeXML::Post::MakeBibliography::NUMBER = 0;
-  local %LaTeXML::Post::MakeBibliography::STYLE =
-    (map { ($_ => $bib->getAttribute($_)) } qw(bibstyle citestyle sort));
+    my $entries = $self->getBibEntries($doc, $bib);
+    # Remove any bibentry's (these should have been converted to bibitems)
+    $doc->removeNodes($doc->findnodes('//ltx:bibentry'));
+    foreach my $biblist ($doc->findnodes('//ltx:biblist')) {
+      $doc->removeNodes($biblist) unless element_nodes($biblist); }
 
-  my $entries = $self->getBibEntries($doc);
-  # Remove any bibentry's (these should have been converted to bibitems)
-  $doc->removeNodes($doc->findnodes('//ltx:bibentry'));
-  foreach my $biblist ($doc->findnodes('//ltx:biblist')) {
-    $doc->removeNodes($biblist) unless element_nodes($biblist); }
-
-  if ($$self{split}) {
-    # Separate by initial.
-    my $split = {};
-    foreach my $sortkey (keys %$entries) {
-      my $entry = $$entries{$sortkey};
-      $$split{ $$entry{initial} }{$sortkey} = $entry; }
-    @bib = map { $self->rescan($_) }
-      $self->makeSubCollectionDocuments($doc, $bib,
-      map { ($_ => $self->makeBibliographyList($doc, $_, $$split{$_})) }
-        sort keys %$split); }
-  else {
-    $doc->addNodes($bib, $self->makeBibliographyList($doc, undef, $entries));
-    @bib = ($self->rescan($doc)); }
-  return @bib; }
+    if ($$self{split}) {
+      # Separate by initial.
+      my $split = {};
+      foreach my $sortkey (keys %$entries) {
+        my $entry = $$entries{$sortkey};
+        $$split{ $$entry{initial} }{$sortkey} = $entry; }
+      @docs = map { $self->rescan($_) }
+        $self->makeSubCollectionDocuments($doc, $bib,
+        map { ($_ => $self->makeBibliographyList($doc, $bib, $_, $$split{$_})) }
+          sort keys %$split); }
+    else {
+      $doc->addNodes($bib, $self->makeBibliographyList($doc, $bib, undef, $entries));
+      #      @docs = ($self->rescan($doc)); }
+      $self->rescan($doc); } }
+  return @docs; }
 
 # Try to preserve the original form & case of the provided Bibkeys
 # HOWEVER, we downcase them before indexing & looking them up!!!!
@@ -99,7 +104,9 @@ sub getBibliographies {
     @bibnames = @{ $$self{bibliographies} }; }
   else {
     my $bibnode = $doc->findnode('//ltx:bibliography');
-    if (my $files = $bibnode->getAttribute('files')) {
+    if (my $files = $bibnode->getAttribute('files')
+      || $bibnode->parentNode->getAttribute('files')    # !!!!!
+    ) {
       @bibnames = map { $_ . '.bib' } split(',', $files); } }
   my @paths = $doc->getSearchPaths;
   my @bibs  = ();
@@ -125,10 +132,11 @@ sub getBibliographies {
       # NOTE: When better integrated with Core, should also check for cached bib documents.
       if (my $xmlpath = pathname_find($bib, paths => [@paths], types => ['xml'])) {
         $bibdoc = $doc->newFromFile($xmlpath); }    # doc will do the searching...
-      elsif (my $bibpath = pathname_find($bib, paths => [@paths], types => ['bib'])) {
+      elsif (my $bibpath = pathname_find($bib, paths => [@paths], types => ['bib'])
+        || pathname_kpsewhich($bib)) {
         $bibdoc = $self->convertBibliography($doc, $bibpath); }
       else {
-        Error('expected', $bib, $self,
+        Error('missing_file', $bib, $self,
           "Couldn't find Bibliography '$bib'",
           "Searchpaths were " . join(',', @paths)); } }
     if ($bibdoc) {
@@ -154,6 +162,13 @@ sub convertBibliography {
   my ($self, $doc, $bib) = @_;
   require LaTeXML;
   require LaTeXML::Common::Config;
+  my @packages =
+    my @preload = ();
+  # Might want/need to preload more (all?) packages, but at least do inputenc!
+  foreach my $po ($self->find_documentclass_and_packages($doc)) {
+    my ($pkg, $options) = @$po;
+    if ($pkg eq 'inputenc') {
+      push(@preload, "[$options]$pkg"); } }
   NoteProgress(" [Converting bibliography $bib ...");
   my $bib_config = LaTeXML::Common::Config->new(
     cache_key      => 'BibTeX',
@@ -162,12 +177,32 @@ sub convertBibliography {
     format         => 'dom',
     whatsin        => 'document',
     whatsout       => 'document',
-    verbosity      => -5,
-    bibliographies => []);
+    bibliographies => [],
+    (@preload ? (preload => [@preload]) : ()));
   my $bib_converter = LaTeXML->get_converter($bib_config);
+  # Tricky and HACKY, we need to release the log to capture the inner workings separately.
+  # ->bind_log analog:
+  my $biblog = '';
+  my $biblog_handle;
+  open($biblog_handle, ">>", \$biblog) or Error("Can't redirect STDERR to log for inner bibliography converter!");
+  *BIB_STDERR_SAVED = *STDERR;
+  *STDERR           = *$biblog_handle;
+  # end ->bind_log
+
   $bib_converter->prepare_session($bib_config);
   my $response = $bib_converter->convert($bib);
-  print STDERR $$response{log};
+
+  # ->flush_log analog:
+  close $biblog_handle;
+  *STDERR = *BIB_STDERR_SAVED;
+  # end ->flush_log
+
+  # Trim log to look internal and report.
+  $biblog =~ s/^.+?\(Digesting/\n\(Digesting/s;
+  $biblog =~ s/Conversion complete:.+$//s;
+  print STDERR $biblog;
+  MergeStatus($$bib_converter{latexml}{state});
+
   # TODO: We need to handle the logging properly, it's a bit of a mess for nested ->convert() calls
   if (my $bibdoc = $$response{result}) {
     NoteProgress("... converted!]");
@@ -187,8 +222,11 @@ sub convertBibliography {
 #  referrers : array of ID's of places that refer to this bibentry
 #  suffix    : a,b... if adjacent author/year are identical.
 
+# NOTE: biblist now have @lists to restrict to include ONLY
+# those bibitems which have been referenced by bibref's for that same list!
+# [Unfortunately, the logic is a bit screwy here]
 sub getBibEntries {
-  my ($self, $doc) = @_;
+  my ($self, $doc, $bib) = @_;
 
   # First, scan the bib files for all ltx:bibentry's, (hash key is bibkey)
   # Also, record the citations from each bibentry to others.
@@ -202,22 +240,31 @@ sub getBibEntries {
       $entries{$bibkey}{citations} = [map { normalizeBibKey($_) } grep { $_ } map { split(',', $_->value) }
           $bibdoc->findnodes('.//@bibrefs', $bibentry)];
       # Also, register the key with the DB, if the bibliography hasn't already been scanned.
+      # NOTE: This associates a SINGLE id with the bibkey
+      # which isn't ideal for multiple bibliographies (see formatBibEntry)
       $$self{db}->register("BIBLABEL:$bibkey", id => $bibid);
-    } }
+  } }
   # Now, collect all bibkeys that were cited in other documents (NOT the bibliography)
   # And note any referrers to them (also only those outside the bib)
   my $citestar = $$self{db}->lookup('BIBLABEL:*');
+  my @lists    = split(/\s+/, $bib->getAttribute('lists') || 'bibliography');
 
   my @queue = ();
   foreach my $dbkey ($$self{db}->getKeys) {
     if ($dbkey =~ /^BIBLABEL:(.*)$/) {
-      my $bibkey = $1;
-      if (my $referrers = $$self{db}->lookup($dbkey)->getValue('referrers')) {
+      my $bibkey     = $1;
+      my $bibdbentry = $$self{db}->lookup($dbkey);
+      my $inlists    = $bibdbentry->getValue('inlist');
+      next if $inlists && !grep { $$inlists{$_} } @lists;
+      if (my $referrers = $bibdbentry->getValue('referrers')) {
         foreach my $refr (keys %$referrers) {
           my ($rid, $e, $t) = ($refr, undef, undef);
           while ($rid && ($e = $$self{db}->lookup("ID:$rid")) && (($t = ($e->getValue('type') || '')) ne 'ltx:bibitem')) {
             $rid = $e->getValue('parent'); }
-          if ($t ne 'ltx:bibitem') {
+          if (!$e) {
+            Warn('expected', 'entry', undef,
+              "Didn't find an entry for reference id=$rid"); }
+          elsif ($t ne 'ltx:bibitem') {
             $entries{$bibkey}{referrers}{$refr} = 1; } }
         push(@queue, $bibkey) if keys %{ $entries{$bibkey}{referrers} }; }
       elsif ($citestar) {    # If \cite{*} include all of them.
@@ -253,7 +300,7 @@ sub getBibEntries {
       my $title = $doc->findnode('ltx:bib-title', $bibentry);
       $date  = ($date  ? $date->textContent  : '');
       $title = ($title ? $title->textContent : '');
-      $$entry{ay} = "$names.$date";
+      $$entry{ay}      = "$names.$date";
       $$entry{initial} = $doc->initial($names, 1);
       # Include this entry keyed using a sortkey.
       $$included{ lc(join('.', $sortnames, $date, $title, $bibkey)) } = $entry;
@@ -309,22 +356,30 @@ sub getNameText {
 # Convert hash of bibentry(s) into biblist of bibitem(s)
 
 sub makeBibliographyList {
-  my ($self, $doc, $initial, $entries) = @_;
-  my $id = $doc->getDocumentElement->getAttribute('xml:id') || 'bib';
+  my ($self, $doc, $bib, $initial, $entries) = @_;
+  my $id = $bib->getAttribute('xml:id')
+    || $doc->getDocumentElement->getAttribute('xml:id') || 'bib';
   $id .= ".L1";
   $id .= ".$initial" if $initial;
   return ['ltx:biblist', { 'xml:id' => $id },
-    map { $self->formatBibEntry($doc, $$entries{$_}) } $doc->unisort(keys %$entries)]; }
+    map { $self->formatBibEntry($doc, $bib, $$entries{$_}) } $doc->unisort(keys %$entries)]; }
 
 # ================================================================================
+# NOTE: With multiple bibliographies, the ID of the bibentry isn't necessarily
+# the ID in the "local" bibliography! (ie. bibentry can be repeated in the document!)
 sub formatBibEntry {
-  my ($self, $doc, $entry) = @_;
+  my ($self, $doc, $bib, $entry) = @_;
   my $bibentry   = $$entry{bibentry};
   my $id         = $bibentry->getAttribute('xml:id');
   my $key        = $bibentry->getAttribute('key');
   my $type       = $bibentry->getAttribute('type');
   my @blockspecs = @{ $FMT_SPEC{$type} || [] };
 
+  # Patch the entry's id in case there are multiple bibs.
+  if (my $bibid = $bib->getAttribute('xml:id')
+    || $doc->getDocumentElement->getAttribute('xml:id') || 'bib') {
+    $id =~ s/^bib//; $id = $bibid . $id; }
+
   local $LaTeXML::Post::MakeBibliography::SUFFIX = $$entry{suffix};
   my $number = ++$LaTeXML::Post::MakeBibliography::NUMBER;
 
@@ -334,33 +389,36 @@ sub formatBibEntry {
   #------------------------------
   # Format the bibtag's
   my @tags = ();
-  push(@tags, ['ltx:bibtag', { role => 'number', class => 'ltx_bib_number' }, $number]);  # number tag
+  push(@tags, ['ltx:tag', { role => 'number', class => 'ltx_bib_number' }, $number]);    # number tag
 
   # Set up authors and fullauthors tags
   my @names = $doc->findnodes('ltx:bib-name[@role="author"]/ltx:surname', $bibentry);
   @names = $doc->findnodes('ltx:bib-name[@role="editor"]/ltx:surname', $bibentry) unless @names;
+  my $etal = 0;
+  if (@names && ($names[-1]->toString eq 'others')) {                                    # Magic!
+    $etal = 1; }
   if (@names > 2) {
-    push(@tags, ['ltx:bibtag', { role => 'authors', class => 'ltx_bib_author' },
+    push(@tags, ['ltx:tag', { role => 'authors', class => 'ltx_bib_author' },
         $doc->cloneNodes($names[0]->childNodes),
         ['ltx:text', { class => 'ltx_bib_etal' }, ' et al.']]);
     my @fnames = ();
     foreach my $n (@names[0 .. $#names - 1]) {
       push(@fnames, $n->childNodes, ', '); }
-    push(@tags, ['ltx:bibtag', { role => 'fullauthors', class => 'ltx_bib_author' },
+    push(@tags, ['ltx:tag', { role => 'fullauthors', class => 'ltx_bib_author' },
         $doc->cloneNodes(@fnames),
         ' and ', $doc->cloneNodes($names[-1]->childNodes)]); }
   elsif (@names > 1) {
-    push(@tags, ['ltx:bibtag', { role => 'authors', class => 'ltx_bib_author' },
+    push(@tags, ['ltx:tag', { role => 'authors', class => 'ltx_bib_author' },
         $doc->cloneNodes($names[0]->childNodes),
         ' and ', $doc->cloneNodes($names[1]->childNodes)]); }
   elsif (@names) {
-    push(@tags, ['ltx:bibtag', { role => 'authors', class => 'ltx_bib_author' },
+    push(@tags, ['ltx:tag', { role => 'authors', class => 'ltx_bib_author' },
         $doc->cloneNodes($names[0]->childNodes)]); }
 
   # Put a key tag, to use in place of authors if needed (esp for software, websites, etc)
   my $keytag;
   if ($keytag = $doc->findnode('ltx:bib-key', $bibentry)) {
-    push(@tags, ['ltx:bibtag', { role => 'key', class => 'ltx_bib_key' },
+    push(@tags, ['ltx:tag', { role => 'key', class => 'ltx_bib_key' },
         $doc->cloneNodes($keytag->childNodes)]); }
 
   my @year = ();
@@ -369,18 +427,18 @@ sub formatBibEntry {
     if (my $datetext = $date->textContent) {
       if ($datetext =~ /^(\d\d\d\d)/) {    # Extract 4 digit year, if any
         @year = ($1); } }
-    push(@tags, ['ltx:bibtag', { role => 'year', class => 'ltx_bib_year' },
+    push(@tags, ['ltx:tag', { role => 'year', class => 'ltx_bib_year' },
         $doc->cloneNodes(@year), ($$entry{suffix} || '')]); }
 
   # Store a type tag, to use in place of year, if needed (esp for software, ...)
   my $typetag;
   if ($typetag = $doc->findnode('ltx:bib-type', $bibentry)) {
-    push(@tags, ['ltx:bibtag', { role => 'bibtype', class => 'ltx_bib_type' },
+    push(@tags, ['ltx:tag', { role => 'bibtype', class => 'ltx_bib_type' },
         $doc->cloneNodes($typetag->childNodes)]); }
 
   # put in the title
   if (my $title = $doc->findnode('ltx:bib-title', $bibentry)) {
-    push(@tags, ['ltx:bibtag', { role => 'title', class => 'ltx_bib_title' },
+    push(@tags, ['ltx:tag', { role => 'title', class => 'ltx_bib_title' },
         $doc->cloneNodes($title->childNodes)]); }
 
   # And finally, the refnum; we need to know the desired citation style!
@@ -389,7 +447,7 @@ sub formatBibEntry {
   my $style = $LaTeXML::Post::MakeBibliography::STYLE{citestyle} || 'numbers';
   $style = 'numbers' unless (@names || $keytag) && (@year || $typetag);
   if ($style eq 'numbers') {
-    push(@tags, ['ltx:bibtag', { role => 'refnum', class => 'ltx_bib_key', open => '[', close => ']' }, $number]); }
+    push(@tags, ['ltx:tag', { role => 'refnum', class => 'ltx_bib_key', open => '[', close => ']' }, $number]); }
   elsif ($style eq 'AY') {
     my @rfnames;
     if (my @authors = $doc->findnodes('ltx:bib-name[@role="author"]/ltx:surname', $bibentry)) {
@@ -407,7 +465,7 @@ sub formatBibEntry {
       $aa = uc(substr($rfnames[0]->textContent, 0, 3)); }
     my $yrtext = (@year ? join('', map { (ref $_ ? $_->textContent : $_); } @year) : '');
     my $yy = (length($yrtext) >= 2 ? substr($yrtext, 2, 2) : $yrtext);
-    push(@tags, ['ltx:bibtag', { role => 'refnum', class => 'ltx_bib_abbrv', open => '[', close => ']' },
+    push(@tags, ['ltx:tag', { role => 'refnum', class => 'ltx_bib_abbrv', open => '[', close => ']' },
         $aa . $yy . ($$entry{suffix} || '')]); }
 
   else {
@@ -421,7 +479,7 @@ sub formatBibEntry {
       @rfnames = $keytag->childNodes; }
     my @rfyear = (@year ? (@year, ($$entry{suffix} || ''))
       : ($typetag ? $typetag->childNodes : ()));
-    push(@tags, ['ltx:bibtag', { role => 'refnum', class => 'ltx_bib_author-year' },
+    push(@tags, ['ltx:tag', { role => 'refnum', class => 'ltx_bib_author-year' },
         $doc->cloneNodes(@rfnames), ' (', $doc->cloneNodes(@rfyear), ')']); }
 
   #------------------------------
@@ -432,7 +490,7 @@ sub formatBibEntry {
     foreach my $row (@$blockspec) {
       my ($xpath, $punct, $pre, $class, $op, $post) = @$row;
       my $negated = $xpath =~ s/^!\s*//;
-      my @nodes = ($xpath eq 'true' ? () : $doc->findnodes($xpath, $bibentry));
+      my @nodes   = ($xpath eq 'true' ? () : $doc->findnodes($xpath, $bibentry));
       next if ($xpath ne 'true') && ($negated ? @nodes : !@nodes);
       push(@x, $punct) if $punct && @x;
       push(@x, $pre) if $pre;
@@ -450,7 +508,7 @@ sub formatBibEntry {
       "Cited by: ", $doc->conjoin(",\n", @citedby), '.']) if @citedby;
 
   return ['ltx:bibitem', { 'xml:id' => $id, key => $key, type => $type, class => "ltx_bib_$type" },
-    @tags,
+    (@tags ? (['ltx:tags', {}, @tags]) : ()),
     @blocks]; }
 
 # ================================================================================
@@ -476,9 +534,19 @@ sub do_name {
 sub do_names {
   my (@names) = @_;
   my @stuff = ();
+  my $sep  = (scalar(@names) > 2 ? ', ' : ' ');
+  my $etal = 0;
+  if (@names && ($names[-1]->textContent eq 'others')) {    # Magic!
+    pop(@names);
+    $etal = 1; }
+  my $n = scalar(@names);
   while (my $name = shift(@names)) {
-    push(@stuff, (@names ? ', ' : ' and ')) if @stuff;
+    if (@stuff) {
+      push(@stuff, $sep);
+      push(@stuff, 'and ') if !$etal && !@names; }
     push(@stuff, do_name($name)); }
+  if ($etal) {
+    push(@stuff, $sep, ['ltx:text', { class => 'ltx_bib_etal' }, 'et al.']); }
   return @stuff; }
 
 sub do_names_short {
@@ -714,4 +782,3 @@ $FMT_SPEC{techreport}            = $FMT_SPEC{report};
 
 # ================================================================================
 1;
-
